@@ -141,7 +141,7 @@ MarketInstructionEnum = Enum(
 )
 
 
-def build_market_instruction(instruction, accounts: list[AccountMeta]) -> Instruction:
+def build_market_instruction(instruction, accounts) -> Instruction:
     return Instruction(
         program_id=get_market_program_id(),
         data=MarketInstructionEnum.build(instruction),
